/****************************************************************
 * music.c
 *---------------------------------------------------------------
 * Copyright (C) 2014-2015 Osman Celimli
 * For conditions of distribution and use, see copyright notice in hscc.c
 ****************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "bytes.h"
#include "useful.h"
#include "hu.h"
#include "sfx.h"
#include "music.h"
#include "notes.h"
/****************************************************************
 * Globals and Externs
 ****************************************************************/
/* Parsing State
 -----------------------------*/
char *pszMusFile = NULL;
char *pszLineBuf = NULL;
unsigned int uiCurLine = 0;
unsigned int uiCurOffset = 0;

unsigned int uiCurChannel = 0;


/* Active Blocks and Tags
 -----------------------------*/
MUS_Block *paMusBlocks = NULL;
unsigned int uiBlocks = 0,uiBlockBufLen = 0;
MUS_Tag *paMusTags = NULL;
unsigned int uiTags = 0,uiTagBufLen = 0;

int aiMainBlock_Found[MUSIC_CHANNELS] = {0};
unsigned int auiMainBlock_Priority[MUSIC_CHANNELS] = {0};
unsigned int auiMainBlock_Offsets[MUSIC_CHANNELS] = {0};

unsigned int uiCurBlock = 0;
unsigned int uiBlockBase = 0;


/* Active Instrument and Note
 -----------------------------*/
unsigned int uiCurInst = 0;
int iInstSet = 0;

int iNoteOn = 0;
int iNoiseInst = 0;
double dInstTuning = 32.0;

/****************************************************************
 * Text Tables
 ****************************************************************/
char REM_MUSMode[]=		"MUS Mode...\n";
char REM_MUSDone[]=		"The current MUS binary is : ";

char ERR_NoMusMem[]=	"ERROR- Out of memory\n";
char ERR_MusRange[]=	"ERROR- Value out of range\n";

char ERR_MusSyntax[]=	"ERROR- Invalid command syntax in : ";
char ERR_MusGarbage[]=	"ERROR- Garbage at end of line in : ";

char ERR_MusBadLabel[]=	"ERROR- Invalid music block label in : ";
char ERR_MusBadPri[]=	"ERROR- Music block priority out of range in : ";
char ERR_MusTwoMains[]=	"ERROR- Second main block declaration in : ";
char ERR_MusTwoLabels[]="ERROR- Block already defined with label : ";
char ERR_MusNoBrace[]=	"ERROR- Expecting '{' at block start : ";
char ERR_MusBlockBusy[]="ERROR- Unexpected end of music block in : ";
char ERR_MusLostBlock[]="ERROR- Unable to resolve reference to block : ";

char ERR_MusBadInst[]=	"ERROR- Instrument not found : ";
char ERR_MusNoNoise_A[]="ERROR- Instrument ";
char ERR_MusNoNoise_B[]=" not compatible with channel ";
char ERR_MusNoiseHelp[]="ERROR- Instruments using noise can only be on channels 5 or 6\n";
char ERR_MusNoInst[]=	"ERROR- Instrument has not been set\n";

char MUS_WhiteSpace[]=	" ,\t\n";

/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----                MUSIC COMMANDS                    --!! *
 * 
 ****************************************************************/
typedef int (*MusCom)(MUS_Block *pBlock);

/* using,u instrument
 -----------------------------*/
int musCom_using(MUS_Block *pBlock)
{
	int iNewInst,iNeedNoise;
	double dTuning;
	char *pszSubTok;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if((iNewInst = sfx_find(pszSubTok)) < 0)
	{
		fprintf(stderr, "%s%s\n", ERR_MusBadInst, pszSubTok);
		return -1;
	}

	uiCurInst = iNewInst;
	if(sfx_getInstInfo(uiCurInst, &iNeedNoise, &dTuning))
		return 1;

	if(iNeedNoise && (HU_LFSR_A != uiCurChannel)
		&& (HU_LFSR_B != uiCurChannel))
	{
		fprintf(stderr, "%s%s%s%u\n", ERR_MusNoNoise_A,
			pszSubTok, ERR_MusNoNoise_B, uiCurChannel);
		fprintf(stderr, "%s", ERR_MusNoiseHelp);
		return 1;
	}

	iInstSet = -1;
	dInstTuning = dTuning;
	iNoiseInst = iNeedNoise;
	return 0;
}

/* priority,pr value
 -----------------------------*/
int musCom_priority(MUS_Block *pBlock)
{
	char *pszSubTok;
	int iStatus,iVal;
	char cPri;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MUSIC_PRIORITY_MIN) || (iVal > MUSIC_PRIORITY_MAX))
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	cPri = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_SET_PRIORITY);
	iStatus |= mus_addByte(pBlock, cPri);
	return iStatus;
}



/* pan,p value
 -----------------------------*/
int musCom_pan(MUS_Block *pBlock)
{
	char *pszSubTok;

	int iStatus,iVal;
	char cPan;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MUSIC_PANNING_MIN) || (iVal > MUSIC_PANNING_MAX))
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	cPan = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_SET_PANNING);
	iStatus |= mus_addByte(pBlock, cPan);
	return iStatus;
}

/* pitch,pt offset [adjustment]
 -----------------------------*/
int musCom_pitch(MUS_Block *pBlock)
{
	char *pszOffTok,*pszAdjTok;
	int16p16_t stOff,stAdj;
	int8p8_t stWrite;
	int iStatus;

	pszOffTok = strtok(NULL, MUS_WhiteSpace);
	pszAdjTok = strtok(NULL, MUS_WhiteSpace);
	if(NULL == pszOffTok)
		return 1;
	if(parse16p8(&stOff, pszOffTok))
		return 1;

	if(NULL != pszAdjTok)
	{
		if(parse16p8(&stAdj, pszAdjTok))
			return 1;
	}
	else
	{
		stAdj.iWhole = 0;
	}

	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_PITCH);

	stWrite.usWhole = stOff.usPair.usHi;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);

	stWrite.usWhole = stAdj.usPair.usHi;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);
	stWrite.usWhole = stAdj.usPair.usLo;
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	return iStatus;
}

/* sample,smp number
 -----------------------------*/
int musCom_sample(MUS_Block *pBlock)
{
	char *pszSubTok;

	int iStatus,iVal;
	char cSample;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MUSIC_SAMPLE_MIN) || (iVal > MUSIC_SAMPLE_MAX))
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	cSample = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_SAMPLE);
	iStatus |= mus_addByte(pBlock, cSample);
	return iStatus;
}

/* sfx,s ticks [offset]
 -----------------------------*/
int musCom_sfx(MUS_Block *pBlock)
{
	char *pszTickTok,*pszOffTok;
	int16p16_t stOff;
	int8p8_t stWrite;

	char cInst,cTicks;
	int iStatus,iVal,iSub;

	if(!iInstSet)
	{
		fprintf(stderr, "%s", ERR_MusNoInst);
		return 1;
	}

	if(NULL == (pszTickTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszTickTok, &iVal))
		return 1;
	if(iVal < MUSIC_WAIT_MIN)
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	pszOffTok = strtok(NULL, MUS_WhiteSpace);
	if(NULL != pszOffTok)
	{
		if(parse16p16(&stOff, pszOffTok))
			return 1;
	}
	else
	{
		stOff.iWhole = 0;
	}

	cInst = uiCurInst;
	stWrite.usWhole = stOff.usPair.usHi;
	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_NOTE_ON);
	iStatus |= mus_addByte(pBlock, cInst);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
	iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);

	/**
	 * Since there's no delay parameter at the end of
	 * a note on command, we'll specify the length in
	 * ticks using wait commands.
	 */
	while(iVal > 0)
	{
		if(iVal > MUSIC_WAIT_MAX)
		{
			cTicks = (char)MUSIC_WAIT_MAX;
			iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_DELAY);
			iStatus |= mus_addByte(pBlock, cTicks);

			iVal -= MUSIC_WAIT_MAX;
		}
		else
		{
			iSub = iVal % (MUSIC_WAIT_MAX + 1);
			cTicks = iSub;

			iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_DELAY);
			iStatus |= mus_addByte(pBlock, cTicks);
			iVal -= iSub;
		}
	}
	return iStatus;
}




/* loop,l count
 -----------------------------*/
int musCom_loop(MUS_Block *pBlock)
{
	char *pszSubTok;
	int iStatus,iVal;
	char cCount;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if((iVal < MUSIC_LOOP_MIN) || (iVal > MUSIC_LOOP_MAX))
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	cCount = iVal & 0xFF;
	iStatus = mus_addByte(pBlock, MUSIC_COMMAND_LOOP_START);
	iStatus |= mus_addByte(pBlock, cCount);
	return iStatus;
}

/* endloop,el
 -----------------------------*/
int musCom_endloop(MUS_Block *pBlock)
{
	return mus_addByte(pBlock, MUSIC_COMMAND_LOOP_END);
}

/* call,c label
 -----------------------------*/
int musCom_call(MUS_Block *pBlock)
{
	char *pszSubTok;
	int iStatus;

	if(mus_addByte(pBlock, MUSIC_COMMAND_CALL))
		return -1;
	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(mus_addCallTag(pszSubTok, uiCurOffset, pszMusFile, uiCurLine))
		return -1;

	iStatus = mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	iStatus |= mus_addByte(pBlock, 0);
	return iStatus;
}

/* return,r
 -----------------------------*/
int musCom_return(MUS_Block *pBlock)
{
	return mus_addByte(pBlock, MUSIC_COMMAND_RETURN);
}

/* break,b
 -----------------------------*/
int musCom_break(MUS_Block *pBlock)
{
	return mus_addByte(pBlock, MUSIC_COMMAND_BREAK);
}




/* wait,w ticks
 -----------------------------*/
int musCom_wait(MUS_Block *pBlock)
{
	char *pszSubTok;
	int iStatus,iVal,iSub;
	char cTicks;

	if(NULL == (pszSubTok = strtok(NULL, MUS_WhiteSpace)))
		return 1;
	if(strtoint(pszSubTok, &iVal))
		return 1;
	if(iVal < MUSIC_WAIT_MIN)
	{
		fprintf(stderr, "%s", ERR_MusRange);
		return 1;
	}

	iStatus = 0;
	while(iVal > 0)
	{
		if(iVal > MUSIC_WAIT_MAX)
		{
			cTicks = (char)MUSIC_WAIT_MAX;
			iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_DELAY);
			iStatus |= mus_addByte(pBlock, cTicks);

			iVal -= MUSIC_WAIT_MAX;
		}
		else
		{
			iSub = iVal % (MUSIC_WAIT_MAX + 1);
			cTicks = iSub;

			iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_DELAY);
			iStatus |= mus_addByte(pBlock, cTicks);
			iVal -= iSub;
		}
	}
	return iStatus;
}

/* rest,r ticks
 -----------------------------*/
int musCom_rest(MUS_Block *pBlock)
{
	int iStatus;

	iStatus = 0;
	if(iNoteOn)
	{
		iNoteOn = 0;
		iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_NOTE_OFF);
	}

	iStatus |= musCom_wait(pBlock);
	return iStatus;
}

/* end,e
 -----------------------------*/
int musCom_end(MUS_Block *pBlock)
{
	int iStatus;

	iStatus = 0;
	if(iNoteOn)
	{
		iNoteOn = 0;
		iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_NOTE_OFF);
	}

	iStatus |= mus_addByte(pBlock, MUSIC_COMMAND_SET_PRIORITY);
	iStatus |= mus_addByte(pBlock, MUSIC_PRIORITY_MIN);
	return iStatus;
}



/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----              COMMAND REFERENCE                   --!! *
 * 
 ****************************************************************/
char *apszMusCom_nameLong[]= {
	"using","priority",
	"pan","pitch","sample","sfx",
	"loop","endloop","call","return","break",
	"wait","rest","end",
	NULL
};

char *apszMusCom_nameShort[]= {
	"u","pr",
	"p","pt","smp","s",
	"l","el","c","rt","b",
	"w","r","e",
	NULL
};

MusCom aMusCom_addr[]= {
	musCom_using,musCom_priority,
	musCom_pan,musCom_pitch,musCom_sample,musCom_sfx,
	musCom_loop,musCom_endloop,musCom_call,musCom_return,musCom_break,
	musCom_wait,musCom_rest,musCom_end,
	NULL
};




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----                BIG BAD PARSER                    --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int mus_addNote(MUS_Block *pBlock, char *pszNote)
 * Attempt to parse the current strtok token as a note and
 * add it to the current music stream.
 ****************************************************************/
int mus_addNote(MUS_Block *pBlock, char *pszNote)
{
	double dFreq,dDiv;
	int8p8_t stWrite;
	char cInst;
	int iStatus;

	iStatus = note_getFreq(pszNote, &dFreq) ? 1 : 0;
	if(!iStatus)
	{
		/**
		 * When playing back a normal waveform the sample
		 * pointer is advanced by one every time the
		 * divider expires. It's clocked at 3.58MHz, so:
		 *
		 *           Freq = 3.58MHz / Div
		 *
		 * Since the waveforms can be anywhere from 1 to 32
		 * samples long, we'll use the tuning factor to
		 * calculate the final divider value:
		 *
		 *        Div = 3.58MHz / (Freq * Tuning)
		 *
		 * The divider is only 12-Bits long, so we have to clamp
		 * our calculations at 0 and 4095.
		 *
		 * Noise / LFSR generation uses a completely different
		 * clocking setup where the 3.58MHz clock is prescaled
		 * by 64 and then divided by a 5-Bit (0 to 31) value:
		 *
		 *      Div = 3.58MHz / (Freq * Tuning * 64)
		 *
		 * Instruments can be created and tuned using noise mode
		 * but from experience with the Lynx, I suggest just
		 * playing them back in SFX mode instead of notes.
		 */
		if(!iInstSet)
		{
			fprintf(stderr, "%s", ERR_MusNoInst);
			return 1;
		}

		dDiv = iNoiseInst
			? HU_AUDFREQ / (dFreq * dInstTuning * 64)
			: HU_AUDFREQ / (dFreq * dInstTuning);
		dDiv = iNoiseInst
			? ((dDiv < 0) ? 0 : ((dDiv > 31) ? 31 : dDiv))
			: ((dDiv < 0) ? 0 : ((dDiv > 4095) ? 4095 : dDiv));
		stWrite.usWhole = (int)dDiv;
		cInst = uiCurInst;

		iStatus = mus_addByte(pBlock, MUSIC_COMMAND_NOTE_ON);
		iStatus |= mus_addByte(pBlock, cInst);
		iStatus |= mus_addByte(pBlock, stWrite.cPair.cLo);
		iStatus |= mus_addByte(pBlock, stWrite.cPair.cHi);

		iStatus |= musCom_wait(pBlock);
		iNoteOn = -1;
	}

	return iStatus;
}
/****************************************************************
 * int mus_parseSASS(FILE *pSASS, char *pszFileName)
 * Parse the SASS music script in the file pSASS whose name
 * is stored in pszFileName. The file must be opened beforehand.
 * The current parsing channel in uiCurChannel will be
 * autoincremented after the entire file has been traversed.
 *
 * If any errors are encountered during parsing, a nonzero
 * value will be returned.
 ****************************************************************/
int mus_parseSASS(FILE *pSASS, char *pszFileName)
{
	MUS_Block CurBlock;
	int iNewPri = 0;

	int iAborted = -1,iStatus = 0;
	int iBlockBusy = 0,iBlockBraced = 0;
	char *pszMainTok = NULL,*pszSubTok = NULL;
	unsigned int uiX;

	/**
	 * Reinitialize select globals (since this will be
	 * called once for each SASS source) and default
	 * the temporary use block.
	 */
	pszMusFile = pszFileName;

	uiCurInst = 0;
	iInstSet = 0;
	iNoteOn = 0;

	if(0 == uiCurChannel)
	{
		fprintf(stdout, "%s", REM_MUSMode);
		paMusBlocks = malloc(sizeof(MUS_Block) * MUSIC_PARSER_BLOCK_LEN);
		paMusTags = malloc(sizeof(MUS_Tag) * MUSIC_PARSER_TAG_LEN);

		if((NULL == paMusBlocks) || (NULL == paMusTags))
		{
			fprintf(stderr, "%s", ERR_NoMusMem);
			return -1;
		}
		uiBlockBufLen = MUSIC_PARSER_BLOCK_LEN;
		uiTagBufLen = MUSIC_PARSER_TAG_LEN;

		memset(aiMainBlock_Found, 0, (sizeof(int) * MUSIC_CHANNELS));
		memset(auiMainBlock_Priority, 0, (sizeof(unsigned int) * MUSIC_CHANNELS));
		memset(auiMainBlock_Offsets, 0, (sizeof(unsigned int) * MUSIC_CHANNELS));
	}

	CurBlock.pacBlockStream = NULL;
	CurBlock.pszBlockName = NULL;
	CurBlock.uiOffset = uiCurOffset;
	CurBlock.uiScriptLen = 0;
	CurBlock.uiBufferLen = 0;
	

	/**
	 * !!-- PARSE START --!!
	 * Allocate our linebuffer then begin traversing
	 * the SASS source line by line.
	 */
	if(NULL == (pszLineBuf = malloc(sizeof(char) * MUSIC_PARSER_LINEBUF_LEN)))
	{
		fprintf(stderr, "%s", ERR_NoMusMem);
		return -1;
	}

	for(uiCurLine = 1; ; uiCurLine++)
	{
		/**
		 * Fetch our current line from the source file, stripping
		 * out any ;comments and tokenizing the remaining characters.
		 */
		if(fgets(pszLineBuf, MUSIC_PARSER_LINEBUF_LEN, pSASS) == NULL)
		{
			iAborted = 0;
			break;
		}
		if(strcut(pszLineBuf, ";", '\0')) break;
		pszMainTok = strtok(pszLineBuf, MUS_WhiteSpace);

		if(iBlockBusy && (NULL != pszMainTok))
		{
			/**
			 * !!-- BLOCK PARSING --!!
			 * -----------------------
			 * Once a block has been opened with a label and
			 * curlybrace {, we can actually process its body.
			 */
			if(iBlockBraced)
			{
				if(0 == strcmp(pszMainTok, "}"))
				{
					/**
					 * Once we've hit our closing curlybrace }, it's time
					 * to add the block to our current data stream.
					 * Afterward the block MUST be reset since its name
					 * and data pointers will still be valid.
					 */
					iBlockBusy = 0;
					iBlockBraced = 0;
					if(mus_addBlock(&CurBlock)) break;

					CurBlock.pacBlockStream = NULL;
					CurBlock.pszBlockName = NULL;
					CurBlock.uiScriptLen = 0;
					CurBlock.uiBufferLen = 0;
				}
				else
				{
					/**
					 * Entries in the block body could be either notes or
					 * a command, we'll check for commands first and
					 * then see if the thing is a note if none of the
					 * commands in our table match up with the main token.
					 */
					for(uiX = 0; (NULL != aMusCom_addr[uiX]); uiX++)
					{
						if((0 == strcmp(apszMusCom_nameLong[uiX], pszMainTok)) ||
						   (0 == strcmp(apszMusCom_nameShort[uiX], pszMainTok)))
						{
							break;
						}
					}

					/**
					 * Both the command handlers and the note parser
					 * return a nonzero value in case if error.
					 * A positive nonzero value indicates a syntax
					 * failure error a negative value indicates an
					 * allocation / system failure.
					 */
					iStatus = (NULL != aMusCom_addr[uiX])
						? (*aMusCom_addr[uiX])(&CurBlock)
						: mus_addNote(&CurBlock, pszMainTok);

					if(iStatus > 0)
					{
						fprintf(stderr, "%s%s,%u\n",
							ERR_MusSyntax, pszFileName, uiCurLine);
						break;
					}
					else if(iStatus < 0)
					{
						break;
					}
				}
			}
			else
			{
				if(0 == strcmp(pszMainTok, "{"))
				{
					iBlockBraced = -1;
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n",
						ERR_MusNoBrace, pszFileName, uiCurLine);
					break;
				}
			}
		}
		else if(NULL != pszMainTok)
		{
			/**
			 * !!-- BLOCK DETECTION --!!
			 * -------------------------
			 * We're not currently parsing a block, so we need
			 * to try and find one with the format: label [priority]
			 */
			pszSubTok = strtok(NULL, MUS_WhiteSpace);
			if(NULL != pszSubTok)
			{
				/**
				 * If we found two tokens, the second one will (or should)
				 * be the priority indicator for this file's main block.
				 */
				if(1 == sscanf(pszSubTok, "%d", &iNewPri))
				{
					if((iNewPri < MUSIC_PRIORITY_MIN) || (iNewPri > MUSIC_PRIORITY_MAX))
					{
						fprintf(stderr, "%s%s,%u\n",
							ERR_MusBadPri, pszFileName, uiCurLine);
						break;
					}
					else
					{
						if(aiMainBlock_Found[uiCurChannel])
						{
							fprintf(stderr, "%s%s,%u\n",
								ERR_MusTwoMains, pszFileName, uiCurLine);
							break;
						}

						aiMainBlock_Found[uiCurChannel] = -1;
						auiMainBlock_Priority[uiCurChannel] = iNewPri;
						auiMainBlock_Offsets[uiCurChannel] = uiCurOffset;
					}
				}
				else
				{
					fprintf(stderr, "%s%s,%u\n",
						ERR_MusBadLabel, pszFileName, uiCurLine);
					break;
				}
			}

			/**
			 * Check to see if a block with this one's label already
			 * exists. If not, we can set up the temporary use block
			 * and switch into block parsing mode for our next line.
			 */
			if(uiBlocks > 0)
			{
				iStatus = 0;
				for(uiX = 0; uiX < uiBlocks; uiX++)
				{
					if(0 == strcmp(pszMainTok, paMusBlocks[uiX].pszBlockName))
					{
						fprintf(stderr, "%s%s in : %s,%u\n",
							ERR_MusTwoLabels, pszMainTok,
							pszFileName, uiCurLine);

						iStatus = -1;
						break;
					}
				}

				if(iStatus) break;
			}

			CurBlock.pacBlockStream = malloc(sizeof(char) * MUSIC_PARSER_STREAM_LEN);
			CurBlock.pszBlockName = malloc(sizeof(char) * (strlen(pszMainTok) + 1));
			CurBlock.uiBufferLen = MUSIC_PARSER_STREAM_LEN;
			CurBlock.uiScriptLen = 0;
			CurBlock.uiOffset = uiCurOffset;

			if((NULL == CurBlock.pacBlockStream) || (NULL == CurBlock.pszBlockName))
			{
				fprintf(stderr, "%s", ERR_NoMusMem);
				break;
			}
			strcpy(CurBlock.pszBlockName, pszMainTok);
			iBlockBusy = -1;
			iBlockBraced = 0;
			iInstSet = 0;
			iNoteOn = 0;
		}

		/**
		 * Regardless of mode, check for garbage at the end
		 * of the current parsing line before advancing.
		 */
		pszSubTok = strtok(NULL, MUS_WhiteSpace);
		if(NULL != pszSubTok)
		{
			fprintf(stderr, "%s%s,%u\n",
				ERR_MusGarbage, pszFileName, uiCurLine);
			break;
		}
	}

	/**
	 * Check to see if any errors were encountered during
	 * parsing, and then free our work buffers (if they
	 * were actually in use).
	 */
	if(iBlockBusy)
	{
		fprintf(stderr, "%s%s,%u\n",
			ERR_MusBlockBusy, pszFileName, uiCurLine);
	}

	if(NULL != pszLineBuf) free(pszLineBuf);
	pszLineBuf = NULL;

	if(NULL != CurBlock.pacBlockStream) free(CurBlock.pacBlockStream);
	if(NULL != CurBlock.pszBlockName) free(CurBlock.pszBlockName);

	uiCurChannel++;
	return (iAborted | iBlockBusy);
}




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----                STINKY LINKY                      --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int mus_linkSASS(char **pacMusic, unsigned int *puiBinLen)
 * Resolve all call tags and build the final music binary,
 * yielding its (malloc'd) base address upon completion in
 * *pacMusic and its size in *puiBinLen.
 * Returns nonzero if anything went wrong.
 ****************************************************************/
int mus_linkSASS(char **ppacMusic, unsigned int *puiBinLen)
{
	char *pacMusic = NULL;
	unsigned int uiStreamBase,uiBinLen;

	char *pacInst = NULL;
	unsigned int uiInstLen,uiInstBase,uiNumInst;

	HuAddress ResAddr;
	unsigned int uiX,uiY,uiZ;

	/**
	 * The final music binary will be arranged in three
	 * distinct segments as detailed below:
	 *	-Header
	 *		Contains the magic word, instrument count
	 *		and offset, along with the track descriptors.
	 *	-Music Script Data
	 *		For all six channels in order of compilation.
	 *	-Instrument Data Block (if any)
	 *		See sfx.c for details.
	 *
	 * We're placing the instrument data block after the
	 * music scripts since all of its address calculations
	 * are relative to itself. So why bother forcing apart
	 * the script and header?
	 */
	uiStreamBase = MUSIC_HEADER_LEN;
	uiBinLen = uiStreamBase;
	for(uiX = 0; uiX < uiBlocks; uiX++)
	{
		uiBinLen += paMusBlocks[uiX].uiScriptLen;
	}

	uiInstBase = uiBinLen;
	if(sfx_buildInst(&pacInst, &uiInstLen, &uiNumInst))
		return -1;
	uiBinLen += uiInstLen;

	fprintf(stdout, "%s%u bytes\n", REM_MUSDone, uiBinLen);
	pacMusic = malloc(sizeof(char) * uiBinLen);
	if(NULL == pacMusic)
	{
		fprintf(stderr, "%s", ERR_NoMusMem);
		return -1;
	}

	/**
	 * Copy in each of our compiled music script blocks starting
	 * at the stream base offset right after the header. After
	 * all have been copied in we can resolve any outstanding
	 * calls within the scripts.
	 */
	uiY = uiStreamBase;
	for(uiX = 0; uiX < uiBlocks; uiX++)
	{
		if(paMusBlocks[uiX].uiScriptLen > 0)
		{
			memcpy(pacMusic + uiY, paMusBlocks[uiX].pacBlockStream,
				(sizeof(char) * paMusBlocks[uiX].uiScriptLen));
			uiY += paMusBlocks[uiX].uiScriptLen;
		}
	}

	for(uiX = 0; uiX < uiTags; uiX++)
	{
		uiZ = paMusTags[uiX].uiOffset + uiStreamBase;
		for(uiY = 0; uiY < uiBlocks; uiY++)
		{
			if(0 == strcmp(paMusTags[uiX].pszBlockName,
				paMusBlocks[uiY].pszBlockName)) break;
		}
		if(uiY == uiBlocks)
		{
			fprintf(stderr, "%s%s in : %s,%u",
				ERR_MusLostBlock,
				paMusTags[uiX].pszBlockName,
				paMusTags[uiX].pszSrcFile,
				paMusTags[uiX].uiLine);

			free(pacMusic);
			return -1;
		}

		ResAddr = hu_calcAddr(paMusBlocks[uiY].uiOffset + uiStreamBase);
		pacMusic[uiZ + HU_LO] = ResAddr.ucLow;
		pacMusic[uiZ + HU_HI] = ResAddr.ucHigh;
		pacMusic[uiZ + HU_BANK] = ResAddr.ucBank;
		pacMusic[uiZ + HU_MAP] = ResAddr.ucMap;
	}

	/**
	 * Now that the channel data has been included and all tags
	 * resolved, we can stamp the header, channel descriptors,
	 * and copy in the instrument data.
	 */
	strcpy(pacMusic + MUSIC_HEADER_OFFSET_MAGIC, MUSIC_MAGICWORD);
	uiZ = MUSIC_HEADER_OFFSET_CHANNELS;
	for(uiX = 0; uiX < MUSIC_CHANNELS; uiX++)
	{
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_MAGIC] = MUSIC_CHANNEL_MAGICWORD;
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_PRIORITY] = auiMainBlock_Priority[uiX];

		ResAddr = hu_calcAddr(auiMainBlock_Offsets[uiX] + uiStreamBase);
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_ADDRL] = ResAddr.ucLow;
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_ADDRH] = ResAddr.ucHigh;
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_ADDRB] = ResAddr.ucBank;
		pacMusic[uiZ + MUSIC_CHANNEL_OFFSET_ADDRM] = ResAddr.ucMap;
		uiZ += MUSIC_CHANNEL_DESC_LEN;
	}

	/**
	 * Copy in the instrument table, and if there's not one just
	 * set the instrument count to zero and fill the offset
	 * with 'NONE' for diagnostic use.
	 */
	if((uiNumInst > 0) && (uiInstLen > 0) && (NULL != pacInst))
	{
		ResAddr = hu_calcAddr(uiInstBase);
		pacMusic[MUSIC_HEADER_OFFSET_INST_COUNT] = uiNumInst;
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRL] = ResAddr.ucLow;
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRH] = ResAddr.ucHigh;
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRB] = ResAddr.ucBank;
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRM] = ResAddr.ucMap;

		memcpy(pacMusic + uiInstBase, pacInst, (sizeof(char) * uiInstLen));
		free(pacInst);
	}
	else
	{
		pacMusic[MUSIC_HEADER_OFFSET_INST_COUNT] = 0;
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRL] = 'N';
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRH] = 'O';
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRB] = 'N';
		pacMusic[MUSIC_HEADER_OFFSET_INST_ADDRM] = 'E';
	}

	*ppacMusic = pacMusic;
	*puiBinLen = uiBinLen;
	return 0;
}




/**!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!**
 * !!----               DATA MANAGEMENT                    --!! *
 * 
 ****************************************************************/
/****************************************************************
 * int mus_addByte(MUS_Block *pBlock, char cByte)
 * Add the given byte to this music block's data stream,
 * returning nonzero if any errors occur.
 ****************************************************************/
int mus_addByte(MUS_Block *pBlock, char cByte)
{
	if(tealloc(&(pBlock->pacBlockStream), &(pBlock->uiBufferLen),
		(pBlock->uiScriptLen + 1), sizeof(char))) return -1;

	pBlock->pacBlockStream[pBlock->uiScriptLen] = cByte;
	pBlock->uiScriptLen++;
	uiCurOffset++;
	return 0;
}
/****************************************************************
 * int mus_addBlock(MUS_Block *pBlock)
 * Adds the music block *pBlock to the current block table,
 * returning nonzero if any errors occur.
 ****************************************************************/
int mus_addBlock(MUS_Block *pBlock)
{
	if(tealloc(&paMusBlocks, &uiBlockBufLen, (uiBlocks + 1),
		sizeof(MUS_Block))) return -1;

	paMusBlocks[uiBlocks].pszBlockName = pBlock->pszBlockName;
	paMusBlocks[uiBlocks].pacBlockStream = pBlock->pacBlockStream;
	paMusBlocks[uiBlocks].uiScriptLen = pBlock->uiScriptLen;
	paMusBlocks[uiBlocks].uiBufferLen = pBlock->uiBufferLen;
	paMusBlocks[uiBlocks].uiOffset = pBlock->uiOffset;
	uiBlocks++;
	return 0;
}
/****************************************************************
 * int mus_addCallTag(char *pszTarget, unsigned int uiOffset,
 *					  char *pszSrcFile, unsigned int uiLine)
 * Add a call reference to pszTarget given the current offset,
 * file, and parsing line, returns nonzero if any errors occur.
 ****************************************************************/
int mus_addCallTag(char *pszTarget, unsigned int uiOffset,
				   char *pszSrcFile, unsigned int uiLine)
{
	if(tealloc(&paMusTags, &uiTagBufLen, (uiTags + 1),
		sizeof(MUS_Tag))) return -1;

	paMusTags[uiTags].pszBlockName =
		malloc(sizeof(char) * (strlen(pszTarget) + 1));
	if(NULL == paMusTags[uiTags].pszBlockName)
	{
		fprintf(stderr, "%s", ERR_NoMusMem);
		return -1;
	}

	paMusTags[uiTags].pszSrcFile =
		malloc(sizeof(char) * (strlen(pszSrcFile) + 1));
	if(NULL == paMusTags[uiTags].pszSrcFile)
	{
		free(paMusTags[uiTags].pszBlockName);
		fprintf(stderr, "%s", ERR_NoMusMem);
		return -1;
	}

	strcpy(paMusTags[uiTags].pszBlockName, pszTarget);
	strcpy(paMusTags[uiTags].pszSrcFile, pszSrcFile);
	paMusTags[uiTags].uiOffset = uiOffset;
	paMusTags[uiTags].uiLine = uiLine;
	uiTags++;
	return 0;
}
/****************************************************************
 * void mus_closeTable(void)
 * Close the currently active music table (blocks, tags, etc).
 * All memory will be released back into the wild and the table's
 * base entries retagged as NULL.
 ****************************************************************/
void mus_closeTable(void)
{
	unsigned int uiX;

	if(NULL != paMusBlocks)
	{
		for(uiX = 0; uiX < uiBlocks; uiX++)
		{
			free(paMusBlocks[uiX].pszBlockName);
			free(paMusBlocks[uiX].pacBlockStream);
		}
		free(paMusBlocks);
	}

	if(NULL != paMusTags)
	{
		for(uiX = 0; uiX < uiTags; uiX++)
		{
			free(paMusTags[uiX].pszBlockName);
			free(paMusTags[uiX].pszSrcFile);
		}
		free(paMusTags);
	}

	paMusBlocks = NULL;
	uiBlocks = 0;
	uiBlockBufLen = 0;

	paMusTags = NULL;
	uiTags = 0;
	uiTagBufLen = 0;
}
