;****************************************************************
;		 MCGenjin PC-Engine Memory Mapper		*
;****************************************************************
; Copyright (C) 2014-2015 Osman Celimli
; For conditions of distribution and use, see copyright notice in HuListen.asm
;----------------------------------------------------------------
; MCGenjin is a memory mapper for the PC-Engine console which
; is implementable on a single Altera EPM7032LC44 CPLD.
; It supports up to 8MB of ROM and internal swapping of the 
; datalines for region control. Two user chipselect lines are 
; given, each with a 256KB mapping area.
;
; The 1MB cartslot is remapped as:
;	$00000-$3FFFF Low 256KB of ROM
;	$40000-$7FFFF Selected 256KB ROM Bank
;	$80000-$BFFFF User CS0
;	$C0000-$FFFFF User CS1
; Writes to even addresses in $00000-$7FFFF set
; mapper region control. Odd addresses set the
; current 256KB bank.
;                            76543210
;    Mapper Region Control : xxxxxxxS
;      0 = Data line order normal
;      1 = Data line order reversed
;                            76543210
;    Mapper Bank Control   : xxxBBBBB
;      BBBBB = A22-A18 of the 256KB bank address
;    Note that while reads from ROM are reversed
;      when data line transposition is enabled, writes
;      to the mapper remain unaltered.
;
; The most common configuration for the mapper is nonvolatile
; memory (FRAM) attached to CS0, and SRAM attached to CS1.
; Note that in the case of slow ROMs used with MCGenjin, their 
; CE and OE should be enabled either at all times or using a20
; to prevent delay from relying on the mapper's enable line.
;****************************************************************
.define MCGENJIN_REGION		$FFFE
.define REGION_SWAPPED		$FF
.define REGION_USWAPPED 	$00

.define MCGENJIN_BANK		$FFFF

.define MCBANK_PAGES		$20
.define FRAM_PAGES		$1
.define EXRAM_PAGES		$10

.define BANK_ROM		$20
.define BANK_FRAM		$40
.define BANK_EXRAM		$60

.define FRAM_Base		$6000
.define EXRAM_Base		$4000

;****************************************************************
;			 Useful Macros				*
;****************************************************************
; MCG_Sel
;	Set the swappable OBJ region (MPR4/5) and the MCGenjin
; Mapper banks to the specified Bank and Map stored in MCG_Bank
; and MCG_Map. The previous map state is pushed to the stack.
; A is trashed, all other registers preserved.
;----------------------------------------------------------------
.macro MCG_Sel
	TMA #MPR4
	PHA
	TMA #MPR5
	PHA
	LDA MCG_LastMap
	PHA

	LDA MCG_Bank
	TAM #MPR4
	INA
	TAM #MPR5

	LDA SysRegion
	beq _SetDirect\@

	PHX
	LDX MCG_Map
	LDA.w MCG_SwapTable,X
	STA MCG_LastMap
	STA MCGENJIN_BANK
	PLX
	bra _SetDone\@
_SetDirect\@
	LDA MCG_Map
	STA MCG_LastMap
	STA MCGENJIN_BANK
_SetDone\@
.endm

; MCG_Res
;	Restore the last OBJ bank settings, A is trashed.
;----------------------------------------------------------------
.macro MCG_Res
	PLA
	STA MCG_LastMap
	STA MCGENJIN_BANK
	PLA
	TAM #MPR5
	PLA
	TAM #MPR4
.endm


; MCG_Sel_Low
;	Set the LOW swappable OBJ region (MPR4) and MCGenjin
; Mapper banks to the specified Bank and Map stored in MCG_Bank
; and MCG_Map. The previous map state is pushed to the stack.
; A is trashed, all other registers preserved.
;----------------------------------------------------------------
.macro MCG_Sel_Low
	TMA #MPR4
	PHA
	LDA MCG_LastMap
	PHA

	LDA MCG_Bank
	TAM #MPR4

	LDA SysRegion
	beq _SetDirect\@

	PHX
	LDX MCG_Map
	LDA.w MCG_SwapTable,X
	STA MCG_LastMap
	STA MCGENJIN_BANK
	PLX
	bra _SetDone\@
_SetDirect\@
	LDA MCG_Map
	STA MCG_LastMap
	STA MCGENJIN_BANK
_SetDone\@
.endm

; MCG_Res_Low
;	Restore the last LOW OBJ bank settings, A is trashed.
;----------------------------------------------------------------
.macro MCG_Res_Low
	PLA
	STA MCG_LastMap
	STA MCGENJIN_BANK
	PLA
	TAM #MPR4
.endm

; MCG_Rel [Base] [Offset] [Destination]
;	Resolve an MCGenjin relative address (LO, HI, BANK, MCMAP)
; where [Destination] = [Base] + [Offset], using the LOW OBJ bank.
; Note that the operands must all be absolute addresses (no ZP).
;----------------------------------------------------------------
.macro MCG_Rel
	clc
	LDA.w \1 + LO
	ADC.w \2 + LO
	STA.w \3 + LO

	LDA.w \1 + HI
	ADC.w \2 + HI
	CMP #>ADDR_OBJ_HI
	bcc _writeHiAddr\@
	SBC #>(ADDR_OBJ_HI - ADDR_OBJ_LO)
	sec
_writeHiAddr\@
	STA.w \3 + HI

	LDA.w \1 + BANK
	ADC.w \2 + BANK
	CMP #(BANK_ROM + MCBANK_PAGES)
	bcc _writeBank\@
	SBC #MCBANK_PAGES
	sec
_writeBank\@
	STA.w \3 + BANK

	LDA.w \1 + MCMAP
	ADC.w \2 + MCMAP
	STA.w \3 + MCMAP
.endm