;****************************************************************
;     HuSFX- Sound Effects and Instrument Script Decoding	*
;****************************************************************
; Copyright (C) 2014-2015 Osman Celimli
; For conditions of distribution and use, see copyright notice in HuSound.asm
;----------------------------------------------------------------
; Sound effects and instruments share a common decoding library,
; allowing for chaotic variations on the traditional ADSR envelope.
; However their storage formats and dispatch procedures are unique,
; primarily because instruments are only started and stopped by
; the HuMusic library and sound effects may be started and stopped
; by the user at any time.
;
; As with the HuMusic library, the data structure for sound effects
; have no alignment restrictions whatsoever.
;
;
; ----- SFX PACKAGES -----
; The current sound effect package's base address is stored
; in HuSFX_PackBase, and must reside in ROM. A well behaved package
; will be formatted as shown below:
; 
; Offset	Contents
; -------	---------
; 0		"S"
; 1		"F"
; 2		"X"
; 3		Number of Sound Effects
; 4+
;  \
;   0-3		Sound Effect Start Addresses (Relative Lo, Hi, Bank, Map)
; (4 * (Number of Sound Effects + 1))
;  \
;   Sound Effects
;
;
; ------- SFX HEADER -------
; Sound effects may be composed of one to six channels of script data
; prefaced by a small header and channel descriptor. Since the PC-Engine
; does not have uniform capabilities accross all its audio channels, each
; channel descriptor must include a "requirements" tag indicating whether
; or not the channel script will need Noise / LFSR mode.
;
; Instruments are limited to only a single channel and thus discard this
; header and instead start with a 4-byte relative (to HuMusic_InstBase)
; address of the instrument's note off portion of the script, which is
; then followed by the data / script itself.
;
; Offset	Contents
; -------	---------
; 0		Priority
; 1		Total Channels (1-6)
; 2+		Channel Script Descriptors:
;  \
;   0			Channel Requirements ($00 = Wave / PCM, $FF = Noise / LFSR)
;   1-4			Start Address (Relative Lo, Hi, Bank, Map)
; (2 + (5 * Total Channels)) 
;  \
;   Sound Effect Data / Script
;
;
; Notable Functions
; -----------------
; -HuSFX_Play:
;	Request playback of the sound effect in A with panning in Y.
; -HuSFX_Stop:
;	Cease playback of any sound effect whose priority is in A.
; -HuSFX_StopAll:
;	Cease playback of all sound effects.
;
; Sound Effect Script Commands
;-----------------------
.define HUSFX_END		0
.define HUSFX_DELAY		1

.define HUSFX_LOOP_START	2
.define HUSFX_LOOP_END		3

.define HUSFX_WAVEFORM		4
.define HUSFX_SAMPLE		5
.define HUSFX_LFSR		6
.define HUSFX_FREQUENCY		7
.define HUSFX_VOLUME		8


;****************************************************************
; HuSFX_Play:							*
;	Request playback of the sound effect whose number is in	*
; A and stereo panning is in Y. Up to HUSOUND_REQLEN requests	*
; can be made per frame, any further requests will be ignored.	*
;****************************************************************
.section "HuSFX_Play" force
HuSFX_Play:
	PHX
	sei
	LDX HuSFX_Reqs		; Make sure there's still room
	CPX #HUSOUND_REQLEN	; in the request list to actually
	beq _abortReq		; fit this new one...

	CMP HuSFX_NumSFX	; ...and that the sound effect
	bcs _abortReq		; requested actually exists.

	STA HuSFX_ReqNum,X	; We'll only store the requested
	TYA			; number and panning for now and
	STA HuSFX_ReqPan,X	; add the respective priorities
	TXA			; during the sort in the driver
	STA HuSFX_ReqSorted,X	; update routine.
	INX
	STX HuSFX_Reqs
_abortReq
	cli
	PLX
	rts
.ends
;****************************************************************
; HuSFX_SortReqs:						*
;	Fetch the priority of each sound effect requested for	*
; playback and then sort all requests by their priorities.	*
;****************************************************************
.section "HuSFX_SortReqs" force
HuSFX_SortReqs:
	LDX HuSFX_Reqs		; Only bother if we actually got
	DEX			; any requests...
	bpl _getPriorities
	rts

_getPriorities			; ---- PRIORITY FETCH ----
	LDA HuSFX_ReqNum,X	; SFX Start Addr = Pack Base + StartAddrs[SFX Num]
	STA HuSFX_DirOffset + LO
	STZ HuSFX_DirOffset + HI
	STZ HuSFX_DirOffset + BANK
	STZ HuSFX_DirOffset + MCMAP
	clc
	ROL HuSFX_DirOffset + LO
	ROL HuSFX_DirOffset + HI
	ROL HuSFX_DirOffset + LO
	ROL HuSFX_DirOffset + HI

	MCG_Rel HuSFX_DirBase HuSFX_DirOffset (HuSound_CurStream + ADDR_RAM)

	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + LO
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + HI
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + BANK
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + MCMAP

	MCG_Rel HuSFX_PackBase HuSFX_DirOffset (HuSound_CurStream + ADDR_RAM)

	jsr HuSound_ReadByte	; Read in the priority and required number of
	STA HuSFX_ReqPri,X	; channels for the sound effect, then save our
	jsr HuSound_ReadByte	; current decoding offset (now at the descriptors)
	STA HuSFX_ReqCha,X	; so the dispatcher can use it later.

	LDA HuSound_CurStream + LO
	STA HuSFX_DispDesc_Lo,X
	LDA HuSound_CurStream + HI
	STA HuSFX_DispDesc_Hi,X
	LDA HuSound_CurStream + BANK
	STA HuSFX_DispDesc_Bank,X
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_DispDesc_Map,X

	DEX
	bmi _sortPriorities
	jmp _getPriorities

_sortPriorities			; ---- PRIORITY SORT ----
	LDX HuSFX_Reqs		; Perform a bubble sort based upon sound
	DEX			; effect priority, where higher priority
				; sound effects will appear HIGHER in the
	LDY HuSFX_ReqSorted,X	; sorted list.
_checkNext
	LDA HuSFX_ReqPri,Y	; Note that ReqSorted contains the SORTED
	DEX			; INDEXES of the requests, the requests
	bmi _sortDone		; themselves never actually move.

	LDY HuSFX_ReqSorted,X	; If the HIGHER entry has a GREATER or
	CMP HuSFX_ReqPri,Y	; EQUAL priority, we're all good and can
	bcs _checkNext		; move on to the next entry.

	TYA			; Otherwise we need to swap the indexes
	INX			; in ReqSorted and start the scan over
	LDY HuSFX_ReqSorted,X	; again.
	STA HuSFX_ReqSorted,X
	DEX
	TYA
	STA HuSFX_ReqSorted,X
	bra _sortPriorities

_sortDone
	rts
.ends
;****************************************************************
; HuSFX_Dispatch:						*
;	Walk through the sound effect playback request list,	*
; dispatching each one using HuSFX_StartSFX.			*
;****************************************************************
.section "HuSFX_Dispatch" force
HuSFX_Dispatch:
	LDY HuSFX_Reqs
	DEY
	bpl _checkReq

	rts
_checkReq
	LDX #(HUSOUND_CHANNELS - 1)
_clearTags
	STZ HuSFX_ReqAlloc,X	; Clear out our allocator tags so
	DEX			; multichannel SFX won't dispatch
	bpl _clearTags		; over themselves.

_checkDescriptor		; After predispatch / sort the DispDecs
	LDX HuSFX_ReqSorted,Y	; should have been set to the beginning
	LDA HuSFX_DispDesc_Lo,X	; of the sound effect's channel descriptors.
	STA HuSound_CurStream + LO
	LDA HuSFX_DispDesc_Hi,X
	STA HuSound_CurStream + HI
	LDA HuSFX_DispDesc_Bank,X
	STA HuSound_CurStream + BANK
	LDA HuSFX_DispDesc_Map,X
	STA HuSound_CurStream + MCMAP

	LDA HuSFX_ReqPri,X
	STA HuSFX_NewPri
	PHX
	CLX
	jsr HuSound_ReadByte	; Check if the channel requires LFSR mode, in which
	beq _findChannel	; case we'll only look at the last two channels.
	LDX #(HUSOUND_CHANNELS - 2)

_findChannel			; ---- CHANNEL ALLOCATOR ----
	LDA HuSFX_ReqAlloc,X	; Once we find a channel which we haven't allocated
	bne _nextChannel	; for this SFX already and has a priority less than
				; or equal to what we're dispatching, it's all good.
	LDA HuSFX_NewPri
	CMP HuSound_Priority,X
	bcs _checkMusic
	bra _nextChannel
_checkMusic
	CMP HuMusic_Priority,X	; Dispatch IFF(New Priority > Current && Music),
	bcs _doDispatch		; This will prevent spaced notes in the music from
				; repeatedly knocking out portions of a sound effect.
_nextChannel
	LDA #$FF
	STA HuSFX_ReqAlloc,X
	INX
	CPX #HUSOUND_CHANNELS
	bne _findChannel

	PLX			; <- If we've failed to find a free allocation slot
	STZ HuSFX_Reqs		; and channel with a lower priority, there is no
	rts			; reason to continue the dispatch operation. Any other
				; SFX still in the requeuest queue will be dropped.
_doDispatch
	LDA #$FF		; ---- CHANNEL DISPATCH ----
	STA HuSFX_ReqAlloc,X	; Read in the next four bytes (containing the start
				; offset relative to the SFX pack base), and save our
	jsr HuSound_ReadByte	; current decode offset over the old DispDesc.
	STA HuSFX_DirOffset + LO
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + HI
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + BANK
	jsr HuSound_ReadByte
	STA HuSFX_DirOffset + MCMAP

	STX HuSFX_NewPri	; If there are any more channels to dispatch for this
	PLX			; sound effect, we'll be able to just continue decoding.
	LDA HuSound_CurStream + LO
	STA HuSFX_DispDesc_Lo,X
	LDA HuSound_CurStream + HI
	STA HuSFX_DispDesc_Hi,X
	LDA HuSound_CurStream + BANK
	STA HuSFX_DispDesc_Bank,X
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_DispDesc_Map,X

	MCG_Rel HuSFX_PackBase HuSFX_DirOffset (HuSound_CurStream + ADDR_RAM)

	PHY
	LDY HuSFX_ReqPan,X
	LDA HuSFX_ReqPri,X
	LDX HuSFX_NewPri
	jsr HuSFX_StartSFX
	PLY

	LDX HuSFX_ReqSorted,Y	; Are there any channels left to dispatch for this
	DEC HuSFX_ReqCha,X	; sound effect in particular?
	beq _sfxDone		; If so, head back to check the descriptor of the
	jmp _checkDescriptor	; next channel.

_sfxDone
	DEY			; Any whole sound effects left to dispatch?
	bmi _dispDone
	jmp _checkDescriptor

_dispDone
	STZ HuSFX_Reqs		; Yikes, let's get out of this cycle munching monster
	rts			; before it eats the rest of the frame!
.ends
;****************************************************************
; HuSFX_Stop:							*
;	Cease playback of the sound effect whose priority is	*
; in A. If multiple sound effects with said priority are active	*
; all of them wil be stopped.					*
;****************************************************************
.section "HuSFX_Stop" force
HuSFX_Stop:
	PHX
	PHY

	LDX #(HUSOUND_CHANNELS - 1)
	sei
_killSFX
	CMP HuSound_Priority,X
	bne _checkNext
	LDY HuSound_isInstrument,X
	bne _checkNext

	STZ HuSound_Priority,X
	CMP HuSample_Priority,X
	bne _checkNext

	sei
	STZ HuSample_Playing,X
	STZ HuSound_noWriteback,X
	STZ HuSample_Priority,X
	cli
_checkNext
	DEX
	bpl _killSFX

	cli
	PLY
	PLX
	rts
.ends
;****************************************************************
; HuSFX_StopAll:						*
;	Cease playback of all active sound effects.		*
;****************************************************************
.section "HuSFX_StopAll" force
HuSFX_StopAll:
	PHX

	LDX #(HUSOUND_CHANNELS - 1)
	sei
_killSFX
	LDA HuSound_isInstrument,X
	bne _checkNext

	LDA HuSound_Priority,X
	CMP HuSample_Priority,X
	bne _noSample

	sei
	STZ HuSample_Playing,X
	STZ HuSound_noWriteback,X
	STZ HuSample_Priority,X
	cli
_noSample
	STZ HuSound_Priority,X
_checkNext
	DEX
	bpl _killSFX

	cli
	PLX
	rts
.ends




;****************************************************************
; HuSFX_StartSFX:						*
;	Begin playback of the sound effect whose script base	*
; address is in HuSound_CurStream over the channel whose	*
; number is in X, with panning in Y and priority in A.		*
; Any instrument or sound effect currently occupying the	*
; channel will be automatically stopped.			*
;****************************************************************
.section "HuSFX_StartSFX" force
HuSFX_StartSFX:
	STA HuSound_Priority,X	; Copy our priority and panning, then
	TYA			; clear out our delays and flag this
	STA HuSound_Panning,X	; channel as occupied by a sound effect.

	STZ HuSFX_Delay,X
	STZ HuSFX_StackLevel,X
	STZ HuSound_isInstrument,X
				; ---- STREAM ADDRESS START ----
				; Luckily the dispatcher already resolved
	LDY.w HuSound_DecOffs,X	; this, so we just have to save it.
	LDA HuSound_CurStream + LO
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_CurStream,Y
	rts
.ends
;****************************************************************
; HuSFX_StartInst:						*
;	Begin playback of the instrument whose script base	*
; address is in HuSound_CurStream over the channel whose	*
; number is in X.						*
; 	Any instrument or sound effect currently occupying the	*
; channel will be automatically stopped.			*
;****************************************************************
.section "HuSFX_StartInst" force
HuSFX_StartInst:
	LDA HuMusic_Priority,X	; Since we're effectively copying
	STA HuSound_Priority,X	; the music track's parameters for
	LDA HuMusic_VolFinal,X	; any instrument setups, we'll
	STA HuSound_Panning,X	; need to bring over its current
				; priority and panning.
	STZ HuSFX_Delay,X
	STZ HuSFX_StackLevel,X
	LDA #$FF
	STA HuSound_isInstrument,X
				; ---- NOTE OFF OFFSET FETCH ----
				; First four stream bytes are our note off
	jsr HuSound_ReadByte	; relative address (to HuMusic_InstBase)...
	STA HuSound_TarStream + LO
	jsr HuSound_ReadByte
	STA HuSound_TarStream + HI
	jsr HuSound_ReadByte
	STA HuSound_TarStream + BANK
	jsr HuSound_ReadByte
	STA HuSound_TarStream + MCMAP
				; ---- STREAM ADDRESS START ----
				; ...which is followed by the instrument
	LDY.w HuSound_DecOffs,X	; stream itself (save it for decoding later).
	LDA HuSound_CurStream + LO
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_CurStream,Y
				; ---- NOTE OFF OFFSET RESOLVE ----
	LDY.w HuSound_DecOffs,X ; HuSFX_NoteOff = HuSound_TarStream + HuMusic_InstBase
	clc			; Where TarStream contains the fetched note off offset.
	LDA HuMusic_InstBase + LO
	ADC HuSound_TarStream + LO
	STA HuSFX_NoteOff,Y
	INY

	LDA HuMusic_InstBase + HI
	ADC HuSound_TarStream + HI
	CMP #>ADDR_OBJ_HI
	bcc _writeHiAddr
	SBC #>(ADDR_OBJ_HI - ADDR_OBJ_LO)
	sec
_writeHiAddr
	STA HuSFX_NoteOff,Y
	INY

	LDA HuMusic_InstBase + BANK
	ADC HuSound_TarStream + BANK
	CMP #(BANK_ROM + MCBANK_PAGES)
	bcc _writeBank
	SBC #MCBANK_PAGES
	sec
_writeBank
	STA HuSFX_NoteOff,Y
	INY

	LDA HuMusic_InstBase + MCMAP
	ADC HuSound_TarStream + MCMAP
	STA HuSFX_NoteOff,Y
	rts
.ends




;****************************************************************
; HuSFX_Init:							*
;	Initialize the instrument and sound effect playback	*
; software, and set up the sound effects package (directory)	*
; at HuSFX_PackBase.						*
;****************************************************************
.section "HuSFX_Init" force
HuSFX_Init:
	STZ HuSFX_NumSFX	; Preset sound effect count to zero.
	LDA HuSFX_PackBase + LO
	ORA HuSFX_PackBase + HI
	ORA HuSFX_PackBase + BANK
	ORA HuSFX_PackBase + MCMAP
	bne _checkPack		; As with all other packages, we'll
	rts			; only investigate the SFX package if
				; its base address is nonzero.
_checkPack
	LDA MCG_LastMap
	PHA
	TMA #MPR4
	PHA

	LDA HuSFX_PackBase + LO
	STA HuSound_CurStream + LO
	LDA HuSFX_PackBase + HI
	STA HuSound_CurStream + HI

	LDA HuSFX_PackBase + MCMAP
	STA HuSound_CurStream + MCMAP

	LDA HuSFX_PackBase + BANK
	CMP #BANK_ROM		; I think it is acceptable to have
	bcc _shoveBank		; sound effects libraries limited to
	bra _checkHeader	; ROM only, so we need the bank
_shoveBank			; bumper here.
	ADC #BANK_ROM

_checkHeader
	STA HuSound_CurStream + BANK

	CLY
_checkLetter
	jsr HuSound_ReadByte	; First three bytes of the header
	CMP _validHeader,Y	; should be "SFX"
	bne _done
	INY
	CPY #3
	bne _checkLetter

	jsr HuSound_ReadByte	; If the header is valid, read in
	STA HuSFX_NumSFX	; the sound effects count and save
				; our current offset as the dir base.
	LDA HuSound_CurStream + LO
	STA HuSFX_DirBase + LO
	LDA HuSound_CurStream + HI
	STA HuSFX_DirBase + HI
	LDA HuSound_CurStream + BANK
	STA HuSFX_DirBase + BANK
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_DirBase + MCMAP

_done
	PLA
	TAM #MPR4
	PLA
	STA MCGENJIN_BANK
	rts

_validHeader	.db "SFX"
.ends




;****************************************************************
; HuSFX_Decode:							*
;	Step through the channel's sound effect or instrument	*
; script to change its waveform, volume, and frequency. Should	*
; be called every driver tick if a channel's priority is > 0.	*
;****************************************************************
.section "HuSFX_Decode" force
HuSFX_Decode:
	LDY.w HuSound_DecOffs,X	; Pretty much the same setup as the
	LDA HuSFX_CurStream,Y	; music decoding routine.
	STA HuSound_CurStream + LO
	INY
	LDA HuSFX_CurStream,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuSFX_CurStream,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuSFX_CurStream,Y
	STA HuSound_CurStream + MCMAP

_checkDelay
	LDA HuSFX_Delay,X	; If decoding is stalled on this channel
	beq _getNextCommand	; (rest or wait between commands), just
				; decrement the wait counter and leave.
	DEC HuSFX_Delay,X
	bra _decodeDone		; <- Exit point for decode process
_getNextCommand
	jsr HuSound_ReadByte	; Fetch our next command byte,
	TAY			; then dispatch its respective handler.
	LDA _sfxCommandLo,Y
	STA HuSound_JumpDest + LO
	LDA _sfxCommandHi,Y
	STA HuSound_JumpDest + HI
	jsr HuSound_VarJump + ADDR_RAM

	bra _checkDelay		; Go back and check the delay value to
_decodeDone			; see if its been set (so we can stop).

	LDY.w HuSound_DecOffs,X
	LDA HuSound_CurStream + LO
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuSFX_CurStream,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_CurStream,Y
	rts

;---- Low Command Table ----------------------------------------
_sfxCommandLo	.db <HuSFX_Com_End,<HuSFX_Com_Del

		.db <HuSFX_Com_SetLoop,<HuSFX_Com_Loop

		.db <HuSFX_Com_Wave,<HuSFX_Com_Sample,<HuSFX_Com_Noise
		.db <HuSFX_Com_Freq,<HuSFX_Com_Vol
;---- High Command Table ---------------------------------------
_sfxCommandHi	.db >HuSFX_Com_End,>HuSFX_Com_Del

		.db >HuSFX_Com_SetLoop,>HuSFX_Com_Loop

		.db >HuSFX_Com_Wave,>HuSFX_Com_Sample,>HuSFX_Com_Noise
		.db >HuSFX_Com_Freq,>HuSFX_Com_Vol
.ends
;****************************************************************
; Sound Effect / Instrument Script Commands			*
;****************************************************************
; HUSFX_END()
; Close the sound effect decoding process and silence the channel.
;----------------------------------------------------------------
.section "HuSFX_Com_End" force
HuSFX_Com_End:
	STZ HuSound_WaveMode,X
	STZ HuSound_isInstrument,X

	STZ HuSound_writeWave,X
	STZ HuSound_writeFreq,X
	STZ HuSound_Priority,X
	PLA
	PLA
	rts
.ends

; HUSFX_DELAY(Delay)
; Delay the decoding of the sound effect script by the given
; count in ticks.
;----------------------------------------------------------------
.section "HuSFX_Com_Del" force
HuSFX_Com_Del:
	jsr HuSound_ReadByte
	STA HuSFX_Delay,X
	rts
.ends

; HUSFX_LOOP_START(Loop Count)
; Set the start of a looping region and define how many counts
; the loop will last. Negative loop counts are considered infinite.
;----------------------------------------------------------------
.section "HuSFX_Com_SetLoop" force
HuSFX_Com_SetLoop:
	jsr HuSound_ReadByte

	clc			; Copy the provided loop count...
	LDA HuSFX_StackLevel,X
	ADC.w HuSound_LoopOffs,X
	TAY
	LDA HuSound_CurByte
	STA HuSFX_LoopCount,Y

	clc			; ...then push our current decoding address
	LDY HuSFX_StackLevel,X
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuSound_CurStream + LO
	STA HuSFX_DecoStack,Y
	INY
	LDA HuSound_CurStream + HI
	STA HuSFX_DecoStack,Y
	INY
	LDA HuSound_CurStream + BANK
	STA HuSFX_DecoStack,Y
	INY
	LDA HuSound_CurStream + MCMAP
	STA HuSFX_DecoStack,Y	; So when we LOOP later, all we have to do
				; is pull this value (next command byte).
	INC HuSFX_StackLevel,X
	rts
.ends

; HUSFX_LOOP_END()
; Decrement the current loop count, branching back to the loop start
; if the loop count is greater than zero or negative (infinite).
;----------------------------------------------------------------
.section "HuSFX_Com_Loop" force
HuSFX_Com_Loop:
	LDA HuSFX_StackLevel,X
	bne _stackNotEmpty	; <- Only attempt to LOOP if our stack has
	rts			; at least ONE entry in it, otherwise abort.

_stackNotEmpty
	clc
	DEA
	ADC.w HuSound_LoopOffs,X
	TAY
	LDA HuSFX_LoopCount,Y
	bmi _restoreAddr	; <- Negative loop counts are infinite / always.
	beq _closeLoop		; <- Counts of zero never loop / act as single-shot.

	LDA HuSFX_LoopCount,Y
	DEA
	STA HuSFX_LoopCount,Y
	bne _restoreAddr
_closeLoop
	DEC HuSFX_StackLevel,X
	rts

_restoreAddr
	LDY HuSFX_StackLevel,X
	DEY
	clc
	LDA.w HuSound_StackOffs,X
	ADC HuSound_DecOffs,Y
	TAY
	LDA HuSFX_DecoStack,Y
	STA HuSound_CurStream + LO
	INY
	LDA HuSFX_DecoStack,Y
	STA HuSound_CurStream + HI
	INY
	LDA HuSFX_DecoStack,Y
	STA HuSound_CurStream + BANK
	INY
	LDA HuSFX_DecoStack,Y
	STA HuSound_CurStream + MCMAP
	rts
.ends

; HUSFX_WAVEFORM(Waveform Data / 32 Bytes of 5-Bit uSamples)
; Put the channel into waveform mode and assign it the 32-Bytes
; of 5-Bit unsigned waveform data included in the stream.
;----------------------------------------------------------------
.section "HuSFX_Com_Wave" force
HuSFX_Com_Wave:
	STZ HuSound_WaveMode,X

	LDA.w HuSound_HasLFSR,X	; If the channel was last using LFSR
	beq _skipNoiseKill	; mode, we need to disable it.

	STX PSG_CHANNEL
	STZ PSG_LFSR

_skipNoiseKill
	PHX
	LDY.w HuSound_WaveOffs,X
	LDX #31
_copyWave
	jsr HuSound_ReadByte
	STA HuSound_WaveMem,Y

	INY
	DEX
	bpl _copyWave

	PLX
	LDA #$FF
	STA HuSound_writeWave,X
	rts
.ends

; HUSFX_SAMPLE(Sample Number)
; Start playback of the specified sample over this channel.
;----------------------------------------------------------------
.section "HuSFX_Com_Sample" force
HuSFX_Com_Sample:
	jsr HuSound_ReadByte	; While this is nearly identical to how HuMusic
	TAY			; performs a sample dispatch, allowing sample
				; playback calls through instruments and sound
	sei			; effects not only creates the opportunity for
	LDA HuSample_Playing,X	; cool sounds but for less need to track
	beq _setupSample	; sample numbers.

	LDA HuSound_Priority,X	; We only want to kick out a currently
	CMP HuSample_Priority,X	; playing sample if our priority is
	bcs _setupSample	; greater than or equal to.
	cli
	rts

_setupSample
	LDA HuSound_Priority,X
	STA HuSample_Priority,X
				; Force the channel into DDA mode
	STX PSG_CHANNEL		; and set the current volume and panning.
	LDA #(PSG_CONTROL_SAMPLE | $1F)
	STA PSG_CONTROL		; We need to preemptively disable IRQs
	LDA HuSound_Panning,X	; (even though they may be re-disabled
	STA PSG_PAN		; in HuSample_Play) to prevent the scenario
				; of the writeback routine triggering right
				; as we switch the channel to DDA, and putting
				; it back into wave or lfsr mode.
	TYA
	PHX
	SXY			; A = Sample #, Y = Channel
	jsr HuSample_Play	; Dispatch using HuSample
	cli
	PLX			; X = Channel (again)
	rts
.ends

; HUSFX_LFSR()
; Put the channel in LFSR / Noise mode.
;----------------------------------------------------------------
.section "HuSFX_Com_Noise" force
HuSFX_Com_Noise:
	LDA.w HuSound_HasLFSR,X	; I have no curiosity in finding out which
	bne _hasLFSR		; dimensional rift will open up when a channel
	rts			; not supporting LFSR mode tries to use it.

_hasLFSR
	LDA #$FF
	STA HuSound_WaveMode,X
	rts
.ends

; HUSFX_FREQUENCY(Frequency Lo, Hi, Frequency Adjustment Lo, Hi, De)
; Set the frequency offset and frequency offset adjustment for
; this channel. The decimal of the frequency offset is cleared.
;----------------------------------------------------------------
.section "HuSFX_Com_Freq" force
HuSFX_Com_Freq:
	jsr HuSound_ReadByte
	STA HuSound_FreqOff_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_FreqOff_Hi,X
	STZ HuSound_FreqOff_De,X

	jsr HuSound_ReadByte
	STA HuSound_FreqOffAdj_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_FreqOffAdj_Hi,X
	jsr HuSound_ReadByte
	STA HuSound_FreqOffAdj_De,X
	rts
.ends

; HUSFX_VOLUME(Volume, Volume Adjustment Lo, De)
; Set the volume and volume adjustment for this channel.
;----------------------------------------------------------------
.section "HuSFX_Com_Vol" force
HuSFX_Com_Vol:
	jsr HuSound_ReadByte
	STA HuSound_Vol_Lo,X
	STZ HuSound_Vol_De,X

	jsr HuSound_ReadByte
	STA HuSound_VolAdj_Lo,X
	jsr HuSound_ReadByte
	STA HuSound_VolAdj_De,X
	rts
.ends
